/*
   NSConcreteFileHandle.m

   Copyright (C) 1995, 1996, 1997 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Ovidiu Predescu <ovidiu@net-community.com>
   Date: May 1997

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <Foundation/common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <sys/socket.h>

enum {
    non_blocking =
#if O_NDELAY
	O_NDELAY
#elif FNDELAY
	FNDELAY
#elif O_NONBLOCK
	O_NONBLOCK
#endif
};

#if HAVE_LIBC_H
# include <libc.h>
#else
# include <unistd.h>
#endif

#include <Foundation/NSData.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSNotificationQueue.h>
#include <Foundation/NSRunLoop.h>
#include <Foundation/NSException.h>
#include <Foundation/exceptions/GeneralExceptions.h>
#include <Foundation/exceptions/NSFileHandleExceptions.h>
#include "NSConcreteFileHandle.h"

@implementation NSConcreteFileHandle

- (void)_determineFileType
{
    struct stat statbuf;

    if (fstat (fd, &statbuf) == -1) {
	NSLog (@"warning: cannot determine the type of file descriptor %d in"
	       @" NSFileHandle %x", fd, self);
	type = NSFileHandleNoType;
    }
    else {
#ifdef S_IFSOCK
	if ((statbuf.st_mode & S_IFSOCK) == statbuf.st_mode)
	    type = NSFileHandleSocket;
	else
#endif
#ifdef S_IFIFO
	/* System V FIFOs, treat them as sockets since the system
	   calls that use them are similar. */
	if ((statbuf.st_mode & S_IFIFO) == statbuf.st_mode)
	    type = NSFileHandleSocket;
	else
#endif
	    type = NSFileHandleUnixFile;

#if 0
	NSLog (@"file type is %s, mode = %o",
	       type == NSFileHandleUnixFile ? "file" : "socket",
	       statbuf.st_mode);
#endif

	/* Is st_blksize set on System V? */
	blockSize = statbuf.st_blksize;
	if (!blockSize)
	    blockSize = 4096;
    }
}

- (id)initWithFileDescriptor:(int)_fd
{
    return [self initWithFileDescriptor:_fd closeOnDealloc:NO];
}

- (id)initWithFileDescriptor:(int)_fd closeOnDealloc:(BOOL)flag
{
    self = [super initWithFileDescriptor:_fd];
    if (!self)
	return nil;

    closeOnDealloc = flag;
    isOpened = YES;
    operation = NSFileHandleNoOperation;
    [self _determineFileType];
    [self setDelegate:self];
    return self;
}

- (id)initWithPath:(NSString*)aPath flags:(int)flags createMode:(int)mode
{
    self = [super initWithPath:aPath flags:flags createMode:mode];
    if (!self)
	return nil;

    closeOnDealloc = YES;
    isOpened = YES;
    operation = NSFileHandleNoOperation;
    [self _determineFileType];
    [self setDelegate:self];
    return self;
}

- (void)dealloc
{
    if (closeOnDealloc && isOpened)
      close (fd);
    [super dealloc];
}

- (NSData*)availableData
{
    NSData* data;

    /* Force the compiler to allocate `data' on stack */
    *&data = nil;

    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		    initWithFileHandle:self
		    operation:@"Try to read from a closed file descriptor!"]);

    if (type == NSFileHandleUnixFile) {
	data = [self readRestOfFile];
    }
    else if (type == NSFileHandleSocket) {
	data = [self readDataOfLength:blockSize];
    }
    else
	THROW([[NSFileHandleUnknownTypeException alloc]
			initWithFileHandle:self]);

    return data;
}

- (NSData*)readDataToEndOfFile
{
    return [self readDataOfLength:UINT_MAX];
}

- (NSData*)readDataOfLength:(unsigned int)length
{
    NSData* data;

    /* Force the compiler to allocate `data' on stack */
    *&data = nil;

    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		    initWithFileHandle:self
		    operation:@"Try to read from a closed file descriptor!"]);

    if (type == NSFileHandleUnixFile) {
	data = [self readFileLength:length];
    }
    else if (type == NSFileHandleSocket) {
	char* buffer;
	unsigned howMany;

	/* Force the compiler to allocate `buffer' on stack */
	*&buffer = Malloc(length);

	howMany = read (fd, buffer, length);
	if (howMany == -1) {
	    Free(buffer);
	    THROW([[NSFileHandleOperationException alloc]
		      initWithFileHandle:self
		      operation:@"Error while reading from socket!"]);
	}
	if (howMany < length)
	    buffer = Realloc(buffer, howMany);
	data = [NSData dataWithBytesNoCopy:buffer length:howMany];
    }
    else
	THROW([[NSFileHandleUnknownTypeException alloc]
			initWithFileHandle:self]);

    return data;
}

- (void)writeData:(NSData*)data
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		    initWithFileHandle:self
		    operation:@"Try to write in a closed file descriptor!"]);

    TRY {
	[super writeData:data];
    } END_TRY
    CATCH(PosixFileOperationException) {
	THROW([[NSFileHandleUnknownTypeException alloc]
			initWithFileHandle:self]);
    } END_CATCH
}

- (void)_registerToRunLoopForModes:(NSArray*)_modes
{
    NSRunLoop* runLoop = [NSRunLoop currentRunLoop];
    int i, count;

    if (!_modes)
	_modes = [NSArray arrayWithObject:[runLoop currentMode]];

    ASSIGN(modes, _modes);

    count = [modes count];
    for (i = 0; i < count; i++)
	[runLoop addPosixFileDescriptor:self forMode:[modes objectAtIndex:i]];
}

- (void)acceptConnectionInBackgroundAndNotifyForModes:(NSArray*)_modes
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Try to accept connections on a closed socket!"]);
    if (operation != NSFileHandleNoOperation) {
	NSString* reason = [NSString stringWithFormat:@"Another operation (%s)"
		      @" is already in progress!",
		      (operation == NSFileHandleAcceptOperation
		       ? "accept"
		       : (operation == NSFileHandleReadOperation
			  ? "read"
			  : (operation == NSFileHandleReadToEndOfFileOperation
			     ? "read to end of file"
			     : "unknown")))];
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:reason]);
    }

    operation = NSFileHandleAcceptOperation;
    fileActivity = NSPosixReadableActivity;

    [self _registerToRunLoopForModes:_modes];
}

- (void)acceptConnectionInBackgroundAndNotify
{
    [self acceptConnectionInBackgroundAndNotifyForModes:nil];
}

- (void)readInBackgroundAndNotifyForModes:(NSArray*)_modes
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Try to read in background from a closed file"
			    @" descriptor!"]);
    if (operation != NSFileHandleNoOperation) {
	NSString* reason = [NSString stringWithFormat:@"Another operation (%s)"
		      @" is already in progress!",
		      (operation == NSFileHandleAcceptOperation
		       ? "accept"
		       : (operation == NSFileHandleReadOperation
			  ? "read"
			  : (operation == NSFileHandleReadToEndOfFileOperation
			     ? "read to end of file"
			     : "unknown")))];
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:reason]);
    }

    operation = NSFileHandleReadOperation;
    fileActivity = NSPosixReadableActivity;

    [self _registerToRunLoopForModes:_modes];
}

- (void)readInBackgroundAndNotify
{
    [self readInBackgroundAndNotifyForModes:nil];
}

- (void)readToEndOfFileInBackgroundAndNotifyForModes:(NSArray*)_modes
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Try to read to end of file in background from a"
			    @" closed file descriptor!"]);
    if (operation != NSFileHandleNoOperation) {
	NSString* reason = [NSString stringWithFormat:@"Another operation (%s)"
		      @" is already in progress!",
		      (operation == NSFileHandleAcceptOperation
		       ? "accept"
		       : (operation == NSFileHandleReadOperation
			  ? "read"
			  : (operation == NSFileHandleReadToEndOfFileOperation
			     ? "read to end of file"
			     : "unknown")))];
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:reason]);
    }

    operation = NSFileHandleReadToEndOfFileOperation;
    fileActivity = NSPosixReadableActivity;

    [self _registerToRunLoopForModes:_modes];
}

- (void)readToEndOfFileInBackgroundAndNotify
{
    [self readToEndOfFileInBackgroundAndNotifyForModes:nil];
}

- (unsigned long long)offsetInFile
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Error getting the offset from a closed file!"]);
    if (type != NSFileHandleUnixFile)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Error getting an offset from a non-file!"]);

    return [self filePosition];
}

- (unsigned long long)seekToEndOfFile
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		initWithFileHandle:self
		operation:@"Error seeking to end of file on a closed file!"]);
    if (type != NSFileHandleUnixFile)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Error seeking to end of file on a non-file!"]);

    [self seekToEnd];
    return [self offsetInFile];
}

- (void)seekToFileOffset:(unsigned long long)offset
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		initWithFileHandle:self
		operation:@"Error seeking to a position on a closed file!"]);
    if (type != NSFileHandleUnixFile)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Error seeking to a position on a non-file!"]);

    [self seekToPosition:offset];
}

- (void)closeFile
{
    if (isOpened)
	close (fd);

    isOpened = NO;
}

- (void)truncateFileAtOffset:(unsigned long long)offset
{
    if (!isOpened)
	THROW([[NSFileHandleOperationException alloc]
		initWithFileHandle:self
		operation:@"Error truncating a closed file!"]);
    if (type != NSFileHandleUnixFile)
	THROW([[NSFileHandleOperationException alloc]
		  initWithFileHandle:self
		  operation:@"Error truncating a non-file!"]);

    [self truncateAtPosition:offset];
}

- (void)activity:(NSPosixFileActivities)activity
  posixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor
{
    NSNotification* notification = nil;
    NSDictionary* userInfo;

    switch (activity) {
	case NSPosixReadableActivity:
	    if (operation == NSFileHandleAcceptOperation) {
		NSFileHandle* newFileHandle;
		union {
		    struct sockaddr_in in_addr;
		    struct sockaddr_un un_addr;
		} sock_address;
		unsigned len = sizeof(sock_address);
		int newFd;

		operation = NSFileHandleNoOperation;
		newFd = accept (fd, (struct sockaddr*)&sock_address, &len);
		newFileHandle = [[[NSFileHandle alloc]
				      initWithFileDescriptor:newFd
					      closeOnDealloc:YES]
				      autorelease];
		userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
			newFileHandle, NSFileHandleNotificationFileHandleItem,
			modes, NSFileHandleNotificationMonitorModes,
			nil];
		notification = [NSNotification notificationWithName:
				  NSFileHandleConnectionAcceptedNotification
						object:self
						userInfo:userInfo];
	    }

	    else if (operation == NSFileHandleReadOperation) {
		/* Try to avoid a blocking operation in case the file
		   descriptor is a socket, so we differentiate here between
		   normal files and sockets. */
		NSData* data = nil;

		if (type == NSFileHandleUnixFile) {
		    /* Reading from files cannot block so do an unconditional
		       read here. */
		    data = [self readDataOfLength:blockSize];
		}
		else if (type == NSFileHandleSocket) {
		    /* Determine if the file descriptor does nonblocking I/O.
		       If it doesn't temporary change it to do, read as many
		       bytes as possible and change it back to blocking I/O.
		     */
		    unsigned fileStatus = fcntl (fd, F_GETFL, 0);
		    BOOL isNonBlocking = (fileStatus & non_blocking);

		    /* Temporary set the file status to nonblocking I/O */
		    if (!isNonBlocking) {
			if(fcntl(fd, F_SETFL, fileStatus|non_blocking) == -1) {
			    NSLog (@"cannot set nonblocking status to file "
				   @"descriptor %d of NSFileHandle %x",
				   fd, self);
			}
		    }

		    data = [self readDataOfLength:blockSize];

		    /* Restore the blocking I/O if it's the case */
		    if (!isNonBlocking) {
			if(fcntl(fd, F_SETFL, fileStatus) == -1) {
			    NSLog (@"cannot restore blocking status of file "
				   @"descriptor %d of NSFileHandle %x",
				   fd, self);
			}
		    }
		}

		userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
			data, NSFileHandleNotificationDataItem,
			modes, NSFileHandleNotificationMonitorModes,
			nil];
		notification = [NSNotification notificationWithName:
				  NSFileHandleReadCompletionNotification
						object:self
						userInfo:userInfo];
	    }

	    else if (operation == NSFileHandleReadToEndOfFileOperation) {
		NSData* data = nil;

		/* We use a similar algorithm as above here to read from
		   the file descriptor. */
		if (type == NSFileHandleUnixFile) {
		    /* Reading from files cannot block so do an unconditional
		       read here. */
		    data = [self readDataToEndOfFile];
		}
		else if (type == NSFileHandleSocket) {
		    unsigned fileStatus = fcntl (fd, F_GETFL, 0);
		    BOOL isNonBlocking = (fileStatus & non_blocking);

		    /* Determine if the file descriptor does nonblocking I/O.
		       If it doesn't temporary change it to do, read as many
		       bytes as possible and change it back to blocking I/O.
		     */
		    /* Temporary set the file status to nonblocking I/O */
		    if (!isNonBlocking) {
			if(fcntl(fd, F_SETFL, fileStatus|non_blocking) == -1) {
			    NSLog (@"cannot set nonblocking status to file "
				   @"descriptor %d of NSFileHandle %x",
				   fd, self);
			}
		    }

		    data = [self readDataToEndOfFile];

		    /* Restore the blocking I/O if it's the case */
		    if (!isNonBlocking) {
			if(fcntl(fd, F_SETFL, fileStatus) == -1) {
			    NSLog (@"cannot restore blocking status of file "
				   @"descriptor %d of NSFileHandle %x",
				   fd, self);
			}
		    }

		    /* Post the notification */
		    userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
			    data, NSFileHandleNotificationDataItem,
			    modes, NSFileHandleNotificationMonitorModes,
			    nil];
		    notification = [NSNotification notificationWithName:
			  NSFileHandleReadToEndOfFileCompletionNotification
						    object:self
						    userInfo:userInfo];
		}
	    }
	    [[NSNotificationCenter defaultCenter]
		    postNotification:notification];
	    [self ceaseMonitoringFileActivity];
	    break;

	default:
	    break;
    }
}

@end /* NSConcreteFileHandle */


@implementation NSNullDeviceFileHandle

- (NSData*)availableData
{
    return [NSData data];
}

- (NSData*)readDataToEndOfFile
{
    return [NSData data];
}

- (NSData*)readDataOfLength:(unsigned int)length
{
    return [NSData data];
}

- (void)writeData:(NSData*)data
{}

- (void)acceptConnectionInBackgroundAndNotifyForModes:(NSArray*)_modes
{}

- (void)acceptConnectionInBackgroundAndNotify
{}

- (void)readInBackgroundAndNotifyForModes:(NSArray*)_modes
{}

- (void)readInBackgroundAndNotify
{}

- (void)readToEndOfFileInBackgroundAndNotifyForModes:(NSArray*)_modes
{}

- (void)readToEndOfFileInBackgroundAndNotify
{}

- (unsigned long long)offsetInFile
{
    return -1;
}

- (unsigned long long)seekToEndOfFile
{
    return -1;
}

- (void)seekToFileOffset:(unsigned long long)offset
{}

- (void)synchronizeFile
{}

- (void)closeFile
{}

- (void)truncateFileAtOffset:(unsigned long long)offset
{}

- (int)fileDescriptor
{
    return -1;
}

@end /* NSNullDeviceFileHandle */
