/* 
   NSHost.m

   Copyright (C) 1995, 1996, 1997 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Aleksandr Savostyanov <sav@conextions.com>
   Author: Ovidiu Predescu <ovidiu@net-community.com>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <config.h>

#ifdef HAVE_LIBC_H
# include <libc.h>
#else
# include <unistd.h>
#endif

#if HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <string.h>

#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <Foundation/NSHost.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSThread.h>

extern NSRecursiveLock* libFoundationLock;

@interface NSHost(Private)
- initWithHostEntity:(struct hostent *)hostEntity;
- initWithName:(NSString*)name;
- initWithAddress:(NSString*)address;
@end

@implementation NSHost

static BOOL cacheEnabled = YES;
static NSMutableDictionary* hostsByName = nil;
static NSMutableDictionary* hostsByAddress = nil;
static NSRecursiveLock* hostsLock = nil;

+ (void)initialize
{
    hostsByName = [NSMutableDictionary new];
    hostsByAddress = [NSMutableDictionary new];
    [[NSNotificationCenter defaultCenter]
	addObserver:self
	selector:@selector(taskNowMultiThreaded:)
	name:NSWillBecomeMultiThreadedNotification
	object:nil];
}

+ (void)taskNowMultiThreaded:(NSNotification*)notification
{
    hostsLock = [NSRecursiveLock new];
}

+ (NSHost *)currentHost
{
    char buffer[1024];
    NSString *hostName;
    gethostname(buffer, 1024);
    hostName = [[NSString alloc] initWithCString:buffer];
    return [NSHost hostWithName:hostName];
}

+ (NSHost *)hostWithName:(NSString *)name
{
    return [[[NSHost alloc] initWithName:name] autorelease];
}

+ (NSHost *)hostWithAddress:(NSString *)address
{
    return [[[NSHost alloc] initWithAddress:address] autorelease];
}

+ (void)setHostCacheEnabled:(BOOL)flag
{
    cacheEnabled = flag;
}

+ (BOOL)isHostCacheEnabled
{
    return cacheEnabled;
}

+ (void)flushHostCache
{
    [hostsByName removeAllObjects];
    [hostsByAddress removeAllObjects];
}

- (void) dealloc
{
    [names release];
    [addresses release];
    [super dealloc];
}

- (BOOL)isEqualToHost:(NSHost *)aHost
{
    NSArray *theAddresses;
    int i, count;

    theAddresses = [aHost addresses];
    count = [theAddresses count];
    for (i = 0; i < count; i++) {
        if ([addresses containsObject:[theAddresses objectAtIndex:i]])
	    return YES;
    }

    return NO;
}

- (BOOL)isEqual:(id)anotherHost
{
    return [self isEqualToHost:anotherHost];
}

- (NSString *)name
{
    if ([names count])
        return [names objectAtIndex:0];
    return nil;
}

- (NSArray *)names
{
    return names;
}

- (NSString *)address
{
    if ([addresses count])
        return [addresses objectAtIndex:0];
    return nil;
}

- (NSArray *)addresses
{
    return addresses;
}

@end

@implementation NSHost(Private)

- initWithName:(NSString*)name
{
    struct hostent *hostEntity;

    if (cacheEnabled) {
	NSHost* host;

	[hostsLock lock];
	host = [hostsByName objectForKey:name];
	[hostsLock unlock];
	if (host)
	    return host;
    }

    [libFoundationLock lock];
    hostEntity = gethostbyname((char*)[name cString]);
    if (!hostEntity) {
        [self release];
        [libFoundationLock unlock];
        return nil;
    }

    [self initWithHostEntity:hostEntity];
    [libFoundationLock unlock];

    return self;
}

- initWithAddress:(NSString*)address
{
    struct hostent *hostEntity;
    struct in_addr ipAddress;

    if (cacheEnabled) {
	NSHost* host;

	[hostsLock lock];
	host = [hostsByAddress objectForKey:address];
	[hostsLock unlock];
	if (host)
	    return host;
    }

    ipAddress.s_addr = inet_addr((char*)[address cString]);
    if (ipAddress.s_addr == (unsigned long)-1) {
        [self release];
        return nil;
    }

    [libFoundationLock lock];
    hostEntity = gethostbyaddr((char*)&ipAddress, sizeof(ipAddress), AF_INET);
    [self initWithHostEntity:hostEntity];
    [libFoundationLock unlock];

    return self;
}

- initWithHostEntity:(struct hostent *)hostEntity
{
    char **ptr;
    struct in_addr ipAddress;

    names = [[NSMutableArray alloc] init];
    [names addObject:[NSString stringWithCString:hostEntity->h_name]];

    ptr = hostEntity->h_aliases;
    while(*ptr) {
        [names addObject:[NSString stringWithCString:*ptr]];
        ptr++;
    }

    addresses = [[NSMutableArray alloc] init];
    ptr = hostEntity->h_addr_list;
    while(*ptr) {
        memcpy(&ipAddress, *ptr, hostEntity->h_length);
        [addresses addObject:
		      [NSString stringWithCString:inet_ntoa(ipAddress)]];
        ptr++;
    }

    if (cacheEnabled) {
	int i, count;

	[hostsLock lock];
	for (i = 0, count = [names count]; i < count; i++)
	    [hostsByName setObject:self forKey:[names objectAtIndex:i]];
	for (i = 0, count = [addresses count]; i < count; i++)
	    [hostsByAddress setObject:self forKey:[addresses objectAtIndex:i]];
	[hostsLock unlock];
    }

    return self;
}

@end
