/* 
   NSObject.h

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Ovidiu Predescu <ovidiu@bx.logicnet.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#ifndef __NSObject_h__
#define __NSObject_h__

#include <objc/objc.h>

#define LIB_FOUNDATION_LIBRARY 1

#define LIB_FOUNDATION_MAJOR_VERSION 0
#define LIB_FOUNDATION_MINOR_VERSION 8
#define LIB_FOUNDATION_SUBMINOR_VERSION 0

#ifndef NSTimeInterval__defined
#define NSTimeInterval__defined
typedef double NSTimeInterval;
#endif

@class Protocol;
@class NSZone;
@class NSString;
@class NSCoder;
@class NSArchiver;
@class NSMethodSignature;
@class NSInvocation;

/*
 * NSObject protocol
 */

@protocol NSObject

/* Identifying Class and Superclass */
- (Class)class;
- (Class)superclass;

/* Determining Allocation Zones */
- (NSZone*)zone;

/* Identifying Proxies */
- (BOOL)isProxy;

/* Testing Inheritance Relationships */
- (BOOL)isKindOfClass:(Class)aClass;
- (BOOL)isMemberOfClass:(Class)aClass;

/* Testing for Protocol Conformance */
- (BOOL)conformsToProtocol:(Protocol*)aProtocol;

/* Testing Class Functionality */
- (BOOL)respondsToSelector:(SEL)aSelector;

/* Managing Reference Counts */
- (id)autorelease;
- (oneway void)release;
- (id)retain;
- (unsigned int)retainCount;

/* Identifying and Comparing Instances */
- (unsigned)hash;
- (BOOL)isEqual:(id)anObject;
- (id)self;

/* Sending Messages Determined at Run Time */
- (id)performSelector:(SEL)aSelector;
- (id)performSelector:(SEL)aSelector withObject:(id)anObject;
- (id)performSelector:(SEL)aSelector withObject:(id)anObject
  withObject:(id)anotherObject;

/* Obsolete methods */
- (id)perform:(SEL)aSelector;
- (id)perform:(SEL)aSelector withObject:(id)anObject;
- (id)perform:(SEL)aSelector withObject:(id)anObject
  withObject:(id)anotherObject;

/* Describing the Object */
- (NSString*)description;

@end /* NSObject */

/* 
 * Copying Objects (deep immutable copy)
 */

@protocol NSCopying
- (id)copyWithZone:(NSZone*)zone;
@end /* NSCopying */

/* 
 * Copying Objects (shallow mutable copy)
 */

@protocol NSMutableCopying
- (id)mutableCopyWithZone:(NSZone*)zone;
@end /* NSCopying */

/* 
 * Coding/Decoding Objects through NSCoder
 */

@protocol NSCoding
- (void)encodeWithCoder:(NSCoder*)aCoder;
- (id)initWithCoder:(NSCoder*)aDecoder;
@end /* NSCoding */

/*
 * NSObject Class - base class for OpenStep Hierarchy
 */

@interface NSObject <NSObject>
{
    Class isa;
}

/* Initializing the Class */
+ (void)initialize;

/* Creating and Destroying Instances */
+ (id)alloc;
+ (id)allocWithZone:(NSZone*)zone;
+ (id)new;
- (void)dealloc;
- (id)init;

/* Testing Class Functionality */
+ (BOOL)instancesRespondToSelector:(SEL)aSelector;

/* Testing Protocol Conformance */
+ (BOOL)conformsToProtocol:(Protocol*)aProtocol;

/* Obtaining Method Information */
+ (IMP)instanceMethodForSelector:(SEL)aSelector;
- (IMP)methodForSelector:(SEL)aSelector;
+ (NSMethodSignature*)instanceMethodSignatureForSelector:(SEL)aSelector;
- (NSMethodSignature*)methodSignatureForSelector:(SEL)aSelector;

/* Posing */
+ (void)poseAsClass:(Class)aClass;

/* Error Handling */
- (void)doesNotRecognizeSelector:(SEL)aSelector;

/* Sending Deferred Messages */
+ (void)cancelPreviousPerformRequestsWithTarget:(id)aTarget
	selector:(SEL)aSelector
	object:(id)anObject;
- (void)performSelector:(SEL)aSelector
	withObject:(id)anObject
	afterDelay:(NSTimeInterval)delay;

/* Forwarding Messages */
- (void)forwardInvocation:(NSInvocation*)anInvocation;

/* Archiving */
- (id)awakeAfterUsingCoder:(NSCoder*)aDecoder;
- (Class)classForArchiver;
- (Class)classForCoder;
- (id)replacementObjectForArchiver:(NSArchiver*)anArchiver;
- (id)replacementObjectForCoder:(NSCoder*)anEncoder;
+ (void)setVersion:(int)version;
+ (int)version;

/* NSCopying/NSMutableCopying shortcuts */
- (id)copy;
- (id)mutableCopy;

/* Obtaining a string representation */
- (NSString*)stringRepresentation;

@end /* NSObject */

@interface NSObject (GNU)
- (Class)transmuteClassTo:(Class)aClassObject;
- subclassResponsibility:(SEL)aSel;
- shouldNotImplement:(SEL)aSel;
- notImplemented:(SEL)aSel;
@end

typedef enum _NSComparisonResult {
    NSOrderedAscending = -1,
    NSOrderedSame = 0,
    NSOrderedDescending = 1
} NSComparisonResult;

enum {NSNotFound = 0x7fffffff};

#include <Foundation/NSZone.h>
#include <Foundation/NSString.h>
#include <Foundation/NSUtilities.h>
#include <Foundation/NSProcessInfo.h>

#endif /* __NSObject_h__ */
