/* 
   NSRunLoop.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Ovidiu Predescu <ovidiu@bx.logicnet.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <Foundation/common.h>

#include <sys/types.h>
#if HAVE_SYS_ERRNO_H
#include <sys/errno.h>
#endif
#include <errno.h>

#if HAVE_SYS_TIME_H
# include <sys/time.h>	/* for struct timeval */
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_MEMORY_H
# include <memory.h>
#endif

#if !HAVE_MEMCPY
# define memcpy(d, s, n)       bcopy((s), (d), (n))
# define memmove(d, s, n)      bcopy((s), (d), (n))
#endif

#if HAVE_LIBC_H
# include <libc.h>
#else
# include <unistd.h>
#endif

#if HAVE_WINDOWS32_SOCKETS_H
# include <windows.h>
# include <Windows32/Sockets.h>
#endif

#include <Foundation/NSRunLoop.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSString.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDate.h>
#include <Foundation/NSTimer.h>
#include <Foundation/NSUtilities.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSException.h>
#include <Foundation/NSPosixFileDescriptor.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSNotificationQueue.h>
#include <Foundation/UnixSignalHandler.h>

#include <extensions/objc-runtime.h>

NSString* NSDefaultRunLoopMode = @"NSDefaultRunLoopMode";
NSString* NSConnectionReplyMode = @"NSConnectionReplyMode";

@interface NSRunLoopFileDescriptorInfo : NSObject
{
    NSPosixFileDescriptor* pfd;
    NSPosixFileActivities activity;
}
+ (NSRunLoopFileDescriptorInfo*)infoWithPosixFileDescriptor:(NSPosixFileDescriptor*)pfd;
- (NSPosixFileDescriptor*)posixFileDescriptor;
- (NSPosixFileActivities)fileActivity;
- (int)fileDescriptor;
@end

@implementation NSRunLoopFileDescriptorInfo
+ (NSRunLoopFileDescriptorInfo*)infoWithPosixFileDescriptor:(NSPosixFileDescriptor*)_pfd
{
    NSRunLoopFileDescriptorInfo* info = [[self alloc] autorelease];

    info->pfd = [_pfd retain];
    info->activity = [_pfd fileActivity];
    return info;
}

- (NSPosixFileDescriptor*)posixFileDescriptor
{
    return pfd;
}

- (NSPosixFileActivities)fileActivity
{
    return activity;
}

- (BOOL)isEqual:(NSRunLoopFileDescriptorInfo*)otherInfo
{
    return [pfd isEqual:otherInfo->pfd];
}

- (int)fileDescriptor
{
    return [pfd fileDescriptor];
}
@end


@interface NSRunLoopTimerInfo : NSObject
{
    NSTimer* timer;
    NSDate* fireDate;
}

+ (NSRunLoopTimerInfo*)infoWithTimer:(NSTimer*)timer;
- (void)recomputeFireDate;
- (NSComparisonResult)compare:(NSRunLoopTimerInfo*)anObject;
- (NSTimer*)timer;
- (NSDate*)fireDate;
@end

@implementation NSRunLoopTimerInfo
+ (NSRunLoopTimerInfo*)infoWithTimer:(NSTimer*)aTimer
{
    NSRunLoopTimerInfo* info = [[self new] autorelease];

    info->timer = [aTimer retain];
    info->fireDate = [[aTimer fireDate] retain];
    return info;
}

- (void)dealloc
{
    [timer release];
    [fireDate release];
    [super dealloc];
}

- (void)recomputeFireDate
{
    if ([timer isValid]) {
	[fireDate release];
	fireDate = [[timer fireDate] retain];
    }
}

- (NSComparisonResult)compare:(NSRunLoopTimerInfo*)anObject
{
    return [fireDate compare:anObject->fireDate];
}

- (NSTimer*)timer			{ return timer; }
- (NSDate*)fireDate			{ return fireDate; }
@end


@interface NSRunLoopActionHolder : NSObject
{
    id target;
    id argument;
    SEL action;
    int order;
}
+ objectWithTarget:(id)target
  argument:(id)argument
  selector:(SEL)action
  order:(int)order;
- (BOOL)isEqual:(id)anotherHolder;
- (void)execute;
@end

@implementation NSRunLoopActionHolder

+ objectWithTarget:(id)_target
  argument:(id)_argument
  selector:(SEL)_action
  order:(int)_order
{
    NSRunLoopActionHolder* holder = [[self alloc] autorelease];

    holder->target = [_target retain];
    holder->argument = [_argument retain];
    holder->action = _action;
    holder->order = _order;

    return holder;
}

- (unsigned)hash
{
  return [(NSObject*)target hash];
}

- (BOOL)isEqual:(NSRunLoopActionHolder*)anotherHolder
{
    return [target isEqual:anotherHolder->target]
	    && [argument isEqual:anotherHolder->argument]
	    && SEL_EQ(action, anotherHolder->action);
}

- (void)execute
{
    [target performSelector:action withObject:argument];
}

- (NSComparisonResult)compare:(NSRunLoopActionHolder*)anotherHolder
{
    return order - anotherHolder->order;
}

@end


@interface NSRunLoopInputManager : NSObject
{
    NSMutableArray*	fileDescriptors;
    NSMutableArray*	timers;
    NSMutableArray*	otherOperations;
}

- (void)addPosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor;
- (void)removePosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor;

- (void)addTimer:(NSTimer*)aTimer;

- (NSMutableArray*)fileDescriptors;
- (NSMutableArray*)timers;

- (void)addOperation:(NSRunLoopActionHolder*)holder;
- (void)removeOperation:(NSRunLoopActionHolder*)holder;
- (void)performAdditionalOperations;
@end


@implementation NSRunLoopInputManager

- init
{
    fileDescriptors = [NSMutableArray new];
    timers = [NSMutableArray new];
    otherOperations = [NSMutableArray new];
    return [super init];
}

- (void)dealloc
{
    [fileDescriptors release];
    [timers release];
    [otherOperations release];
    [super dealloc];
}

- (void)addPosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor
{
    id info = [NSRunLoopFileDescriptorInfo
		  infoWithPosixFileDescriptor:fileDescriptor];
    [fileDescriptors addObject:info];
}

- (void)removePosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor
{
    id info = [NSRunLoopFileDescriptorInfo
		  infoWithPosixFileDescriptor:fileDescriptor];
    [fileDescriptors removeObject:info];
}

- (void)addTimer:(NSTimer*)aTimer
{
    [timers addObject:[NSRunLoopTimerInfo infoWithTimer:aTimer]];
}

- (void)addOperation:(NSRunLoopActionHolder*)holder
{
    [otherOperations addObject:holder];
    [otherOperations sortUsingSelector:@selector(compare:)];
}

- (void)removeOperation:(NSRunLoopActionHolder*)holder
{
    [otherOperations removeObject:holder];
}

- (void)performAdditionalOperations
{
    [otherOperations makeObjectsPerform:@selector(execute)];
}

- (NSMutableArray*)fileDescriptors	{ return fileDescriptors; }
- (NSMutableArray*)timers		{ return timers; }

@end /* NSRunLoopInputManager */


@implementation NSRunLoop

extern NSRecursiveLock* libFoundationLock;

/* Class variable */
static NSMutableDictionary* runLoopsDictionary = nil;
static NSRunLoop* currentRunLoop = nil;
static BOOL taskIsMultithreaded = NO;

+ (void)initialize
{
    [[NSNotificationCenter defaultCenter]
	addObserver:self
	selector:@selector(taskNowMultiThreaded:)
	name:NSWillBecomeMultiThreadedNotification
	object:nil];
}

+ (NSRunLoop*)currentRunLoop
{
    if (taskIsMultithreaded) {
	NSRunLoop* currentLoop;
	NSThread* currentThread;

	[libFoundationLock lock];

	currentThread = [NSThread currentThread];
	currentLoop = [runLoopsDictionary objectForKey:currentThread];
	if (!currentLoop) {
	    currentLoop = [[self new] autorelease];
	    [runLoopsDictionary setObject:currentLoop forKey:currentThread];
	}

	[libFoundationLock unlock];

	return currentLoop;
    }
    else {
	if (!currentRunLoop)
	    currentRunLoop = [self new];
	return currentRunLoop;
    }
}

+ (void)taskNowMultiThreaded:notification
{
    taskIsMultithreaded = YES;
    runLoopsDictionary = [NSMutableDictionary new];

    if (currentRunLoop) {
	NSThread* currentThread = [NSThread currentThread];

	[runLoopsDictionary setObject:currentRunLoop forKey:currentThread];
	[currentRunLoop release];
	currentRunLoop = nil;
    }
}

- init
{
    self = [super init];
    inputsForMode = [NSMutableDictionary new];
    mode = [NSDefaultRunLoopMode retain];
    return self;
}

- (void)dealloc
{
    [inputsForMode release];
    [mode release];
    [super dealloc];
}

- (NSString*)currentMode
{
    return mode;
}

static int compare_fire_dates(id timer1, id timer2, void* userData)
{
    return [[timer1 fireDate] compare:[timer2 fireDate]];
}

- (NSDate*)limitDateForMode:(NSString*)aMode
{
    NSString* format = @"Caught exception %@ with reason %@ ";
    NSMutableArray* timers = [[inputsForMode objectForKey:aMode] timers];
    volatile int i, count;
    NSMutableArray* copyOfTimers;

    ASSIGN(mode, aMode);

    /* Remove invalid timers */
    for(count = [timers count], i = count - 1; i >= 0; i--)
        if(![[[timers objectAtIndex:i] timer] isValid]) {
            [timers removeObjectAtIndex:i];
	    count--;
        }

    /* Currently only timers have limit dates associated with them */
    if(!count)
        return nil;

    copyOfTimers = [timers mutableCopy];

    /* Sort the timers based on their fire date */
    [copyOfTimers sortUsingFunction:compare_fire_dates context:NULL];

    /* Fire all the timers with their fire date expired */
    for(i = 0; i < count; i++) {
        NSRunLoopTimerInfo* timerInfo = [copyOfTimers objectAtIndex:i];
        NSDate* fireDate = [timerInfo fireDate];
        NSDate* currentDate = [NSDate date];

        if([fireDate earlierDate:currentDate] == fireDate
	   || [fireDate isEqualToDate:currentDate]) {
            NSTimer* timer = [timerInfo timer];
            TRY {
	      [timer fire];
	    } END_TRY
	    OTHERWISE {
	      NSLog(format, [localException name], [localException reason]);
	    } END_CATCH
	    
            if(![timer repeats])
                [timer invalidate];
        }
    }

    [copyOfTimers release];

    /* Recompute the fire dates for this cycle */
    [timers makeObjectsPerform:@selector(recomputeFireDate)];

    /* Sort the timers based on their fire date */
    [timers sortUsingFunction:compare_fire_dates context:NULL];

    return [timers count] ? [[timers objectAtIndex:0] fireDate] : nil;
}

- (void)addTimer:(NSTimer*)aTimer
	forMode:(NSString*)aMode
{
    id manager = [inputsForMode objectForKey:aMode];

    if(!manager) {
	manager = [[NSRunLoopInputManager new] autorelease];
	[inputsForMode setObject:manager forKey:aMode];
    }

    [manager addTimer:aTimer];
}

- (BOOL)runMode:(NSString*)aMode
	beforeDate:(NSDate*)limitDate
{
    id inputManager, fileDescriptors;
    NSAutoreleasePool* pool;
    NSArray* timers;
    NSDate* date;

    /* Retain the limitDate so it doesn't get released by limitDateForMode:
	if it fires a timer that has as fireDate the limitDate.
	(bug report from Benhur Stein <Benhur-de-Oliveira.Stein@imag.fr>)
      */
    [limitDate retain];

    inputManager = [inputsForMode objectForKey:aMode];
    timers = [inputManager timers];
    fileDescriptors = [inputManager fileDescriptors];

    if (([fileDescriptors count] != 0) || ([timers count] != 0)) {
	pool = [NSAutoreleasePool new];
	date = [self limitDateForMode:aMode];
	date = date ? [date earlierDate:limitDate] : limitDate;
	[self acceptInputForMode:aMode beforeDate:date];
	[pool release];
	[limitDate release];
	return YES;
    }

    [limitDate release];
    return NO;
}

/*  Runs the loop until limitDate or until the earliest limit date for input
    sources in the specified mode. */
- (void)acceptInputForMode:(NSString*)aMode
	beforeDate:(NSDate*)limitDate
{
    id inputManager, fileDescriptors;
    struct timeval tp = { 0, 0 };
    struct timeval* timeout = NULL;
    NSTimeInterval delay = 0;
    id fileDescriptorsCopy;
    fd_set readSet, writeSet, exceptionsSet;
    volatile int i, r, count;

    ASSIGN(mode, aMode);

    if(!limitDate)
	limitDate = [NSDate distantFuture];
    else {
	delay = [limitDate timeIntervalSinceNow];
	    /* delay > 0 means a future date */

	/* If limitDate is in the past return */
	if(delay < 0)
	    return;
    }

    inputManager = [inputsForMode objectForKey:aMode];
    fileDescriptors = [inputManager fileDescriptors];
    count = [fileDescriptors count];

    /* Compute the timeout for select */
    if([limitDate isEqual:[NSDate distantFuture]])
	timeout = NULL;
    else {
	tp.tv_sec = delay;
	tp.tv_usec = (delay - (NSTimeInterval)tp.tv_sec) * 1000000.0;
	timeout = &tp;
    }

    ASSIGN(mode, aMode);

    FD_ZERO(&readSet);
    FD_ZERO(&writeSet);
    FD_ZERO(&exceptionsSet);
    for (i = 0; i < count; i++) {
	NSRunLoopFileDescriptorInfo* fileDescriptor
	    = [fileDescriptors objectAtIndex:i];
	unsigned int fileActivity = [fileDescriptor fileActivity];
	int fd = [fileDescriptor fileDescriptor];

	if (fileActivity & NSPosixReadableActivity)
	    FD_SET(fd, &readSet);
	if (fileActivity & NSPosixWritableActivity)
	    FD_SET(fd, &writeSet);
	if (fileActivity & NSPosixExceptionalActivity)
	    FD_SET(fd, &exceptionsSet);
    }

    errno = 0;
    do {
	r = select(FD_SETSIZE, &readSet, &writeSet, &exceptionsSet, timeout);
	if (r == -1) {
	    if (errno == EINTR) {
		/* Interrupt occured; break the loop to give a chance to
		   UnixSignalHandler to handle the signals. */
		errno = 0;
		break;
	    }
	    else
		NSLog (@"select() error ingnored: '%s'", strerror (errno));
	    errno = 0;
	}
    } while (r == -1);

    if(r > 0) {
	NSString* format = @"Caught exception %@ with reason %@ ";

	fileDescriptorsCopy = [fileDescriptors copy];
	TRY {
	  for (i = 0; i < count && r; i++) {
	      NSPosixFileDescriptor* fileDescriptor
		  = [[fileDescriptorsCopy objectAtIndex:i]
			posixFileDescriptor];
	      int fd = [fileDescriptor fileDescriptor];
  
	      if (FD_ISSET(fd, &readSet)) {
		  [[fileDescriptor delegate] activity:NSPosixReadableActivity
		      posixFileDescriptor:fileDescriptor];
		  r--;
	      }
  
	      if (FD_ISSET(fd, &writeSet)) {
		  [[fileDescriptor delegate] activity:NSPosixWritableActivity
		      posixFileDescriptor:fileDescriptor];
		  r--;
	      }
  
	      if (FD_ISSET(fd, &exceptionsSet)) {
		  [[fileDescriptor delegate]
			activity:NSPosixExceptionalActivity
			posixFileDescriptor:fileDescriptor];
		  r--;
	      }
	  }
	} END_TRY
	OTHERWISE {
	  NSLog(format, [localException name], [localException reason]);
	} END_CATCH
	[fileDescriptorsCopy release];
    }

    [inputManager performAdditionalOperations];
    [NSNotificationQueue runLoopASAP];
}

- (void)runUntilDate:(NSDate*)limitDate
{
    BOOL shouldContinue = YES;

    if(!limitDate)
	limitDate = [NSDate distantFuture];
    else {
	/* If limitDate is in the past return */
	if([limitDate timeIntervalSinceNow] < 0)
	    return;
    }

    while (shouldContinue) {
        NSAutoreleasePool* pool = [NSAutoreleasePool new];
        if ([limitDate laterDate:[NSDate date]] == limitDate) {
	    if([self runMode:NSDefaultRunLoopMode beforeDate:limitDate] == NO)
		shouldContinue = NO;
        }
	else
	    shouldContinue = NO;
        [pool release];
    }
}

- (void)run
{
    [self runUntilDate:[NSDate distantFuture]];
}

- (void)performSelector:(SEL)aSelector
  target:(id)target
  argument:(id)anArgument
  order:(unsigned)order
  modes:(NSArray*)modes
{
    id holder = [NSRunLoopActionHolder objectWithTarget:target
					argument:anArgument
					selector:aSelector
					order:order];
    int i, count = [modes count];

    for (i = 0; i < count; i++)
	[[inputsForMode objectForKey:[modes objectAtIndex:i]]
	    addOperation:holder];
}

- (void)cancelPerformSelector:(SEL)aSelector
  target:(id)target
  argument:(id)anArgument
{
    id holder = [NSRunLoopActionHolder objectWithTarget:target
					argument:anArgument
					selector:aSelector
					order:0];
    id enumerator = [inputsForMode keyEnumerator];
    id aMode;

    while ((aMode = [enumerator nextObject]))
	[[inputsForMode objectForKey:aMode] removeOperation:holder];
}

- (void)addPosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor
  forMode:(NSString*)aMode
{
    NSRunLoopInputManager* inputManager = [inputsForMode objectForKey:aMode];

    if(!inputManager) {
	inputManager = [[NSRunLoopInputManager new] autorelease];
	[inputsForMode setObject:inputManager forKey:aMode];
    }
    [inputManager addPosixFileDescriptor:fileDescriptor];
}

- (void)removePosixFileDescriptor:(NSPosixFileDescriptor*)fileDescriptor
  forMode:(NSString*)aMode
{
    NSRunLoopInputManager* inputManager = [inputsForMode objectForKey:aMode];

    if(!inputManager) {
	inputManager = [[NSRunLoopInputManager new] autorelease];
	[inputsForMode setObject:inputManager forKey:aMode];
    }
    [inputManager removePosixFileDescriptor:fileDescriptor];
}

@end /* NSRunLoop */
