/* 
   NSThread.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <config.h>

#include <time.h>

#if HAVE_SYS_TIME_H
# include <sys/time.h>	/* for struct timeval */
#endif

#if HAVE_LIBC_H
# include <libc.h>
#else
# include <unistd.h>
#endif

#include <Foundation/common.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSString.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSUtilities.h>
#include <Foundation/NSObjCRuntime.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSDate.h>
#include <Foundation/NSRunLoop.h>
#include <Foundation/NSNotificationQueue.h>

#include <Foundation/NSException.h>
#include <Foundation/exceptions/GeneralExceptions.h>

#include <extensions/objc-runtime.h>

/*
    NSAutoreleasePool must be changed to return a default pool from 
    the current thread autorelease stack and to operate on that stack
    when pools are allocated and deallocated. The overhead is one more
    method call in +[NSAutoreleasePool defaultPool] to 
    +[NSThread currentThread].

    NSExceptions && exception handlers must be changed to use current
    thread top-level handler instead of internal static variable. The
    overhead is not important since handler are not pushed/extracted
    too often.
 */

/* NSThread notifications */

NSString* NSWillBecomeMultiThreadedNotification = 
	@"NSWillBecomeMultiThreadedNotification";
NSString* NSThreadWillExitNotification =
	@"NSThreadWillExitNotification";

/* Global thread variables */

static BOOL isMultiThreaded = NO;
static NSThread* mainThread = nil;

@implementation NSThread

/*
 * Instance Methods
 */

- initWithTarget:(id)aTarget selector:(SEL)aSelector argument:(id)anArgument
{
    // Not yet running
    isRunning = NO;
    // Set running parameters
    target = [aTarget retain];
    selector = aSelector;
    arg = [anArgument retain];
    // Autorelease pool is nil
    autoreleaseStack = nil;
    // No exception handler
    exceptionStack = NULL;
    // Thread dictionary
    threadDictionary = [NSMutableDictionary new];
    
    return self;
}

- (void)run
{
    id pool;
    
    // Set current thread data
    objc_thread_set_data(self);
    // Make thread-top-level autorelease
    pool = [[NSAutoreleasePool alloc] init];
    runLoop = [NSRunLoop new];
    // Run
    [target perform:selector withObject:arg];
    // pool cleanup
    [pool release];
}

- (void)exit
{
    // Clear autorelease for current thread
    while (autoreleaseStack)
	[autoreleaseStack release];
    objc_thread_exit();
}

- (NSMutableDictionary*)threadDictionary
{
    return threadDictionary;
}

- (void)dealloc
{
    if (isRunning)
	THROW([[InvalidUseOfMethodException alloc] initWithFormat:
	    @"cannot deallocate NSThread for running thread"]);
    [target release];
    [arg release];
    [threadDictionary release];
    [defaultQueue release];
    [runLoop release];
    [super dealloc];
}

- (NSAutoreleasePool*)threadDefaultAutoreleasePool
{
    return autoreleaseStack;
}

- (void)setThreadDefaultAutoreleasePool:(NSAutoreleasePool*)pool
{
    autoreleaseStack = pool;
}

- (void*)threadDefaultExceptionHandler
{
    return exceptionStack;
}

- (void)setThreadDefaultExceptionHandler:(void*)handler
{
    exceptionStack = handler;
}

- (NSRunLoop*)threadRunLoop
{
    return runLoop;
}

- (void)setThreadRunLoop:(NSRunLoop*)aLoop
{
    runLoop = [aLoop retain];
}

- (struct _InstanceList*)threadNotificationQueues
{
    return notificationQueues;
}

- (void)setThreadNotificationQueues:(struct _InstanceList*)theQueues
{
    notificationQueues = theQueues;
}

- (NSNotificationQueue*)defaultNotificationQueue
{
    if (!defaultQueue)
        defaultQueue = [[NSNotificationQueue alloc] init];
    return defaultQueue;
}

- (void)setDefaultNotificationQueue:(NSNotificationQueue*)aQueue
{
    defaultQueue = [aQueue retain];
}

- (NSNotificationCenter*)defaultNotificationCenter
{
    if (!defaultCenter)
        defaultCenter = [[NSNotificationCenter alloc] init];
    return defaultCenter;
}

- (void)setDefaultNotificationCenter:(NSNotificationCenter*)center
{
    defaultCenter = [center retain];
}

/*
 * Class Methods
 */

+ (void)initialize
{
    // We are still in non-multithread state
    if (!mainThread) {
	mainThread = [NSThread alloc];
	mainThread->threadDictionary = [NSMutableDictionary new];
	mainThread->autoreleaseStack = nil;
	mainThread->exceptionStack = NULL;
	objc_thread_set_data(mainThread);
    }
}

#if HAVE_OBJC_THREAD_CREATE
static void nsThreadStartThread(id thread)
{
    [thread run];
//    [thread exit];
    [thread release];
}
#else
+ (void)_runThread:(NSThread*)thread
{
    [thread run];
//    [thread exit];
    [thread release];
}
#endif

+ (void)detachNewThreadSelector:(SEL)aSelector 
  toTarget:(id)aTarget withObject:(id)anArgument
{
    id thread = [[self alloc] 
	initWithTarget:aTarget selector:aSelector argument:anArgument];
    
    if (!isMultiThreaded) {
	[NSAutoreleasePool taskNowMultiThreaded:nil];
	[NSException taskNowMultiThreaded:nil];
	[NSRunLoop taskNowMultiThreaded:nil];
	[NSNotificationQueue taskNowMultiThreaded:nil];
	[NSNotificationCenter taskNowMultiThreaded:nil];
	[[NSNotificationCenter defaultCenter]
	    postNotificationName:NSWillBecomeMultiThreadedNotification
	    object:nil userInfo:nil];
	isMultiThreaded = YES;
    }
    
#if HAVE_OBJC_THREAD_CREATE
    objc_thread_create((void(*)(void*arg))nsThreadStartThread, thread);
#else
    objc_thread_detach(@selector(_runThread:), self, thread);
#endif
}

+ (NSThread*)currentThread
{
    return isMultiThreaded ? (NSThread*)objc_thread_get_data() : mainThread;
}

+ (void)exit
{
    [[NSNotificationCenter defaultCenter]
	postNotificationName:NSThreadWillExitNotification
	object:self userInfo:nil];
    [[self currentThread] exit];
}

+ (BOOL)isMultiThreaded
{
    return isMultiThreaded;
}

+ (void)sleepUntilDate:(NSDate*)aDate
{
    struct timeval tp = { 0, 0 };
    struct timeval* timeout = NULL;
    fd_set set;
    NSTimeInterval delay = 0;

    if (!aDate) return;
    if([aDate earlierDate:[NSDate date]] == aDate) return;

    delay = [aDate timeIntervalSinceNow];
    tp.tv_sec = delay;
    tp.tv_usec = (delay - (NSTimeInterval)tp.tv_sec) * 1000000.0;
    timeout = &tp;

    FD_ZERO(&set);
    select(FD_SETSIZE, &set, NULL, NULL, timeout);
}

@end /* NSThread */
